#include "OptVertexBuffer.h"
#include <vector>
#include <deque>
#include <iostream>
#include <cstring>

using namespace std;

/**
 * Not sure why MSFT linker seems to require this
 */
OptVertexBuffer::~OptVertexBuffer()
{}

/**
 * Fence only used by NVVertexBuffer, no-op by default
 */
void OptVertexBuffer::map()
{}

/**
 * Similarly, no-op by default
 */
void OptVertexBuffer::unmap()
{
}

class DefaultVertexBuffer : public OptVertexBuffer
{
public :
	DefaultVertexBuffer( GLvoid* vertices)
		: m_vertices( (char*)vertices)
	{
	}

	~DefaultVertexBuffer()
	{
	}

	void setVertexPointer( GLint count, GLenum type, GLsizei stride, GLsizei offset)
	{
		glVertexPointer( count, type, stride, m_vertices+offset);
	}
	void setNormalPointer( GLenum type, GLsizei stride, GLsizei offset)
	{
		glNormalPointer( type, stride, m_vertices+offset);
	}
	void setTexCoordPointer( GLint count, GLenum type, GLsizei stride, GLsizei offset)
	{
		glTexCoordPointer( count, type, stride, m_vertices+offset);
	}

private :
	char* m_vertices;
};

class DefaultVertexBufferFactory : public OptVertexBufferFactory
{
public :
	OptVertexBuffer* getBuffer( GLvoid* vertices, GLsizei size)
	{
		return new DefaultVertexBuffer( vertices);
	}

	void release()
	{
	}
};

class ATIVertexBuffer : public OptVertexBuffer
{
public :
	ATIVertexBuffer( GLsizei size, GLvoid* vertices)
	{
		m_objectID = glNewObjectBufferATI( size, vertices, GL_STATIC_ATI);
	}

	~ATIVertexBuffer()
	{
		glDeleteObjectBufferATI( m_objectID);
	}

	void setVertexPointer( GLint count, GLenum type, GLsizei stride, GLsizei offset)
	{
		glArrayObjectATI( GL_VERTEX_ARRAY, count, type, stride, m_objectID, offset);
	}
	void setNormalPointer( GLenum type, GLsizei stride, GLsizei offset)
	{
		glArrayObjectATI( GL_NORMAL_ARRAY, 3, type, stride, m_objectID, offset);
	}
	void setTexCoordPointer( GLint count, GLenum type, GLsizei stride, GLsizei offset)
	{
		glArrayObjectATI( GL_TEXTURE_COORD_ARRAY, count, type, stride, m_objectID, offset);
	}

private :
	GLuint m_objectID;
};

class ATIVertexBufferFactory : public OptVertexBufferFactory
{
public :
	OptVertexBuffer* getBuffer( GLvoid* vertices, GLsizei size)
	{
		return new ATIVertexBuffer( size, vertices);
	}

	void release()
	{

	}
};

class NVVertexBufferFactory : public OptVertexBufferFactory
{
    friend class NVVertexBuffer;

public :
	static NVVertexBufferFactory* m_factory;
	NVVertexBufferFactory( GLsizei typical_size, GLsizei max_total)
	{
		m_typical=typical_size;
        if ( max_total<typical_size)
            max_total=typical_size;
        m_segments=max_total/typical_size;
        if ( max_total % m_segments)
            max_total=m_segments*typical_size;
		m_arena_size=max_total;
		// Allocate video memory
		cerr<<"Trying to allocate memory\n";
		m_video_allocated=false;
		m_arena=(char*)wglAllocateMemoryNV( max_total, 0.0f, 0.0f, 1.0f);
		if ( ! m_arena)
		{
			cerr<<"video failed--  trying for AGP\n";
			m_arena=(char*)wglAllocateMemoryNV( max_total, 0.0f, 0.0f, 0.70f);
		}
		if ( ! m_arena)
		{
			cerr<<"wglAllocate failed\n";
		}
		else
		{
			m_video_allocated=true;
			m_fences=new GLuint[m_segments];
			m_used.assign( m_segments, false);
			glGenFencesNV( m_segments, m_fences);
			glVertexArrayRangeNV( max_total, m_arena);
			glEnableClientState(GL_VERTEX_ARRAY_RANGE_NV);
		}
	}

	OptVertexBuffer* getBuffer( GLvoid* vertices, GLsizei size);

	void release()
	{
		delete m_factory;
		m_factory=0;
	}

private :
	~NVVertexBufferFactory()
	{
		if ( m_video_allocated)
		{
			glDisableClientState(GL_VERTEX_ARRAY_RANGE_NV);
			wglFreeMemoryNV( m_arena);
			glDeleteFencesNV( m_segments, m_fences);
			delete[] m_fences;
		}
	}

	void releaseBuffer( NVVertexBuffer* to_release);
	void freeSegments( NVVertexBuffer* to_free);
	void map( NVVertexBuffer* to_map);


	int m_segments;
	GLsizei m_typical;
	GLsizei m_arena_size;
    GLuint* m_fences;
	char* m_arena;
	bool m_video_allocated;
	vector<bool> m_used;
	deque<NVVertexBuffer*> m_mapped_buffers;
};

NVVertexBufferFactory* NVVertexBufferFactory::m_factory=0;

class NVVertexBuffer : public OptVertexBuffer
{
    friend class NVVertexBufferFactory;
public :
	NVVertexBuffer( NVVertexBufferFactory& factory, int segments, int size, GLvoid* data)
		: m_vertices(0), m_start_segment( -1), m_fence(0), m_factory( factory), m_segments( segments), m_size( size), m_data( data)
	{
	}

	~NVVertexBuffer()
	{
		if ( m_vertices)
			m_factory.releaseBuffer( this);
	}

	void setVertexPointer( GLint count, GLenum type, GLsizei stride, GLsizei offset)
	{
		glVertexPointer( count, type, stride, m_vertices+offset);
	}
	void setNormalPointer( GLenum type, GLsizei stride, GLsizei offset)
	{
		glNormalPointer( type, stride, m_vertices+offset);
	}
	void setTexCoordPointer( GLint count, GLenum type, GLsizei stride, GLsizei offset)
	{
		glTexCoordPointer( count, type, stride, m_vertices+offset);
	}

	void map()
	{
		if ( ! m_vertices)
		{
			m_factory.map( this);
		}
	}

	void unmap()
	{
		glSetFenceNV( m_fence, GL_ALL_COMPLETED_NV);
	}

private :
	/** Only set when mapped */
	char* m_vertices;
	/** Only set when mapped */
	int m_start_segment;
	/** Only set when mapped */
	GLuint m_fence;
	
	NVVertexBufferFactory& m_factory;
	int m_segments;
	GLsizei m_size;
	GLvoid* m_data;
};

OptVertexBuffer* NVVertexBufferFactory::getBuffer( GLvoid* vertices, GLsizei size)
{
    if ( size>m_arena_size || ! m_video_allocated)
        return new DefaultVertexBuffer( vertices);
	int segments=size/m_typical + ( ( size % m_typical) ? 1 :0);
	if ( segments<=0)
		return 0;
	return new NVVertexBuffer( *this, segments, size, vertices);
}

void NVVertexBufferFactory::freeSegments( NVVertexBuffer* to_free)
{
	glFinishFenceNV( to_free->m_fence);
	for ( int i=0; i<to_free->m_segments; i++)
	{
		m_used[i+to_free->m_start_segment]=false;
	}
	to_free->m_vertices=0;
}

void NVVertexBufferFactory::releaseBuffer( NVVertexBuffer* to_release)
{
	for ( deque<NVVertexBuffer*>::iterator i=m_mapped_buffers.begin();
		i!=m_mapped_buffers.end(); i++)
	{
		if ( *i==to_release)
		{
			m_mapped_buffers.erase( i, i+1);
			break;
		}
	}
	freeSegments( to_release);
}

void NVVertexBufferFactory::map( NVVertexBuffer* to_map)
{
	while ( true)
	{
		int found_count=0;
		for ( int i=0; i<m_segments; i++)
		{
			if ( ! m_used[i])
			{
				found_count++;
				if ( found_count==to_map->m_segments)
				{
					for ( int j=i-to_map->m_segments+1; j<=i; j++)
						m_used[j]=true;
					m_mapped_buffers.push_back( to_map);
					to_map->m_vertices=m_arena+i*m_typical;
					memcpy( to_map->m_vertices, to_map->m_data, to_map->m_size);
					to_map->m_start_segment=i;
					to_map->m_fence=m_fences[i];
					return;
				}
			}
			else
				found_count=0;
		}
		NVVertexBuffer* to_free=m_mapped_buffers.front();
		m_mapped_buffers.pop_front();
		freeSegments( to_free);
	}
}

OptVertexBufferFactory& OptVertexBufferFactory::getFactory( GLsizei typical_size,
														   GLsizei max_total)
{
	static ATIVertexBufferFactory ati_factory;
	static DefaultVertexBufferFactory default_factory;
	if ( GL_ATI_vertex_array_object)
		return ati_factory;
	else if ( GL_NV_vertex_array_range)
	{
		if ( ! NVVertexBufferFactory::m_factory)
			NVVertexBufferFactory::m_factory=new NVVertexBufferFactory( typical_size,
				max_total);
		return *NVVertexBufferFactory::m_factory;
	}
	else
		return default_factory;
}
